from mathbox.statistics.estimator import pcc


def max_corr(obj, set, num=1, lag=20):
    ccfs = []
    for i in range(len(set)):
        pcc_list = pcc(obj, set[i], lag)
        ccfs.append((i, max(pcc_list, key=lambda x: x[1])))
    sorted_res = sorted(ccfs, key=lambda d: d[1][1], reverse=True)
    return sorted_res[:num]



import pytest
from ..filter import *

def test_moving_average():
    assert moving_average([1,3,5,4,4],3) == [1.0,2.0,3.0,4.0,pytest.approx(4.3333,0.001)]
    
def test_moving_median():
    assert moving_median([1,2,5,9,6],3) == [1.0,1.5,2.0,5.0,6.0]


from math import pi, cos, sin

# >> Discrete Fourier transform for sampled signals
# x [in]: sampled signals, a list of magnitudes (real numbers)
# y [out]: a list of complex numbers
def dft(x):
    N, y = len(x), []
    for k in range(N):
        real, imag = 0, 0
        for n in range(N):
            theta = -k * (2 * pi) * (float(n) / N)
            real += x[n] * cos(theta)
            imag += x[n] * sin(theta)
        y.append(complex(real, imag))
    return y # 第一个是直流分量

def rdft(x):
    return dft(x)[:(len(x)//2)+1]


# >> Inverse discrete Fourier transform
# yr [in]: real parts of the sinusoids
# yi [in]: imaginary parts of the sinusoids
# x [out]: sampled signals (real parts only), a list of magnitude
def idft(yr, yi):
    N, x = len(yr), []
    for n in range(N):
        real, imag = 0, 0
        for k in range(N):
            theta = k * (2 * pi) * (float(n) / N)
            real += (yr[k] * cos(theta)) - (yi[k] * sin(theta))
            # imag += (yr[k] * math.sin(theta)) + (yi[k] * math.cos(theta))
        x.append(real / N)
    return x

def irdft(fourier, fix=False):
    ext_f = []
    if abs(fourier[-1].imag) and not fix < 1e-10:
        for i in range(len(fourier)-2):
            ext_f.append(fourier[-i-2].conjugate())
    else:
        for i in range(len(fourier)-1):
            ext_f.append(fourier[-i-1].conjugate())
    fourier.extend(ext_f)
    N, x = len(fourier), []
    for n in range(N):
        real, imag = 0, 0
        for k in range(N):
            theta = k * (2 * pi) * (float(n) / N)
            real += (fourier[k].real * cos(theta)) - (fourier[k].imag * sin(theta))
            # imag += (fourier[k].real * math.sin(theta)) + (fourier[k].imag * math.cos(theta))
        x.append(real / N)
    return x

def dftfreq(n, dt = 1.0):
    if not isinstance(n, int):
        raise ValueError("n should be an integer")
    val = 1.0 / (n * dt)
    results = []
    N = (n-1)//2 + 1
    p1 = [x for x in range(0, N)]
    results[:N] = p1
    p2 = [x for x in range(-(n//2), 0)]
    results[N:] = p2
    return [x * val for x in results]


def mean(samples):
    if len(samples) == 0:
        raise ValueError("samples is empty")
    return sum(samples) / len(samples)

"""
ddof == 1 provides an unbiased estimator of the variance of the infinite population.
ddof == 0 provides a maximum likelihood estimate of the variance for normally distributed variables.
"""
def var(samples, ddof=1):
    if ddof != 1 and ddof != 0:
        raise ValueError("ddof must be 0 or 1")
    x_bar = mean(samples)
    if ddof == 1 and len(samples) == 1:
        raise ValueError("unbiased variance requires at least two data points")
    return sum([(x - x_bar) ** 2 for x in samples]) / (len(samples) - ddof)

def std(samples, ddof=1):
    if ddof != 1 and ddof != 0:
        raise ValueError("ddof must be 0 or 1")
    return var(samples, ddof=ddof) ** 0.5


def median(samples) :
    samples = sorted(samples)
    if len(samples) % 2 == 0:
        return (samples[len(samples) // 2 - 1] + samples[len(samples) // 2]) / 2
    else:
        return samples[len(samples) // 2]

def pearson_correlation_coefficient(x, y):
    if len(x) != len(y):
        raise ValueError("x and y must have the same length")
    if max(x) == min(x) or max(y) == min(y):
        raise ValueError("x and y must not be constant")
    x_mean = mean(x)
    y_mean = mean(y)
    x_qsum  = 0
    y_qsum  = 0
    xy_sum  = 0
    for i in range(len(x)):
        x_qsum += (x[i] - x_mean) ** 2
        y_qsum += (y[i] - y_mean) ** 2
        xy_sum += (x[i] - x_mean) * (y[i] - y_mean)
    return xy_sum / (x_qsum * y_qsum) ** 0.5

def pcc(x, y, lag_max=20):
    i = len(x)
    if i != len(y):
        raise ValueError("x and y must have the same length")
    if max(x) == min(x) or max(y) == min(y):
        raise ValueError("x and y must not be constant")
    if lag_max >= i:
        lag_max = i - 1
    result = []
    new_x = [0 for _ in range(i - 1)] + x + [0 for _ in range(i - 1)]
    for lag in range(i - lag_max - 1, i + lag_max):
        x_lag = new_x[lag:lag+i]
        if max(x_lag) == min(x_lag):
            continue
        result.append((lag + 1 - i, pearson_correlation_coefficient(y, x_lag)))
    return result

# TODO:
# ncc not same length, constant???
def _ncc(x,y):
    x_mean = mean(x)
    y_mean = mean(y)
    x_qsum  = 0
    y_qsum  = 0
    xy_sum  = 0
    for i in range(len(x)):
        x_qsum += (x[i] - x_mean) ** 2 + 1e-10
        y_qsum += (y[i] - y_mean) ** 2 + 1e-10
        xy_sum += (x[i] - x_mean) * (y[i] - y_mean)
    return xy_sum / (x_qsum * y_qsum) ** 0.5

def ncc(x,y,lag_max=20):
    i = len(x)
    j = len(y)
    if lag_max >= min(i,j):
        lag_max = min(i,j) - 1
    ncc_list = []
    result = []
    # can be optimized by i > j or i < j
    new_x = [0 for _ in range(j - 1)] + x + [0 for _ in range(j - 1)]
    for lag in range(i + j -1):
        ncc_list.append((lag + 1 - i, _ncc(y, new_x[lag:lag+j])))
    for lag in range(i - lag_max - 1, i + lag_max):
        result.append(ncc_list[lag])
    return result
    

def _operator_4_e_dist(a,b):
    # distance[i * len_a + j(0~len_b)]
    len_a = len(a)
    len_b = len(b)
    distance = []
    sum = 0
    for i in range(len_a):
        row_distance = []
        for j in range(len_b):
            row_distance.append(abs(a[i]-b[j]))
            sum += abs(a[i]-b[j])
        distance.append(row_distance)
    return distance, sum / (len_a * len_b)

def energy_distance(a, b, normalize=False):
    _,A = _operator_4_e_dist(a,b)
    _,B = _operator_4_e_dist(a,a)
    _,C = _operator_4_e_dist(b,b)
    e_distance = (2* A - B - C)
    normailized_e_distance = e_distance / (2 * A + 1e-10)
    if normalize:
        return normailized_e_distance
    else:
        return e_distance




from pytest import approx

from ..interpolation import *


def test_point_linear_interp():
    sig_1 = (0, 1)
    sig_2 = (1, 2)
    assert point_linear_interp(1.5, sig_1[0], sig_1[1], sig_2[0], sig_2[1]) == 2.5
    assert point_linear_interp(.5, sig_1[0], sig_1[1], sig_2[0], sig_2[1]) == 1.5

def test_series_linear_interp():
    orig_time = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    orig_signal = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    desired_time = [-2, -1, 0, 1, 2, 3, 5, 6, 8, 9, 10]
    assert series_linear_interp(orig_time, orig_signal, desired_time, "linear") == desired_time
    assert series_linear_interp(orig_time, orig_signal, desired_time) == [0, 0, 0, 1, 2, 3, 5, 6, 8, 9, 9]


from mathbox.statistics.estimator import energy_distance
from mathbox.statistics.hypothesis_testing import permutation_test
from mathbox.optimizer.utils import argmax
from mathbox.utils import pairwise

def t_stats(a, b):
    e_distance = energy_distance(a,b)
    return e_distance * len(a) * len(b) / (len(a) + len(b))

def _calculate_t_stats(series, pvalue, permutations):
    t_value = []
    #t_value = [0] # magic
    for i in range(len(series)-1):
        a = series[:i+1]
        b = series[i+1:]
        t = t_stats(a,b)
        t_value.append(t)
    idx = int(argmax(t_value))
    a = series[:idx+1]
    b = series[idx+1:]
    if permutation_test(a, b, t_stats, pvalue, permutations):
        return (idx, t_value[idx])
    else:
        return None

def _get_next_significant_change_point(series, change_points, memo, pvalue: float, permutations: int):
    windows = [0] + sorted(change_points) + [len(series)]
    candidates = []
    for bounds in pairwise(windows):
        if bounds in memo:
            candidates.append(memo[bounds])
        else:
            a, b = bounds
            candidate = _calculate_t_stats(series[a:b], pvalue, permutations)
            if candidate is not None:
                #idx = int(argmax(stats))    # 9,8,7 #idx = int(np.argmax(stats)) # 9,0,0
                new = (candidate[0] + a, candidate[1]) # [a,b)里最大的t和它的位置
                memo[bounds] = new
                return new
            else:
                None

def e_divisive(series, jump = 5, pvalue: float = 0.05, permutations: int = 100):
    if jump != int(jump) or jump < 1:
        raise ValueError("jump must be a positive integer")
    change_points = []
    memo = {}
    while significant_change_point := _get_next_significant_change_point(
        series, change_points, memo, pvalue, permutations
    ):
        if significant_change_point is not None:
            if change_points == []:
                change_points.append(significant_change_point[0])
            else:
                result = sorted([abs(significant_change_point[0] - x) for x in change_points])
                if result[0] > jump and significant_change_point[0] > jump and significant_change_point[0] < len(series) - jump:
                    change_points.append(significant_change_point[0])
    return change_points



import random


def permutation_test(alist, blist, estimator, p_value=0.05, times=100):
    total = alist + blist
    len_a = len(alist)
    t_stats = estimator(alist, blist)
    better_num = 0
    for _ in range(times):
        random.shuffle(total)
        new_a = total[:len_a]
        new_b = total[len_a:]
        new_t_stats = estimator(new_a, new_b)
        if new_t_stats >= t_stats:
            better_num += 1
    probability = better_num / (times + 1)
    return True if probability <= p_value else False # false means reject H0


from mathbox.statistics.estimator import mean, std

def noise_outlier(noise, bias=3):
    noise_mean = mean(noise)
    noise_std = std(noise)
    outlier_lo = [(i,x) for i,x in enumerate(noise) if x < noise_mean - bias * noise_std]
    outlier_hi = [(i,x) for i,x in enumerate(noise) if x > noise_mean + bias * noise_std]
    return outlier_lo, outlier_hi

def simple_outlier(series, bias=3):
    sorted_series = sorted(series)
    length = len(series)
    q3 = sorted_series[int(length * 0.75)]
    q1 = sorted_series[int(length * 0.25)]
    outlier_lo = [(i,x) for i,x in enumerate(series) if x < q1 - bias * (q3 - q1)]
    outlier_hi = [(i,x) for i,x in enumerate(series) if x > q3 + bias * (q3 - q1)]
    return outlier_lo, outlier_hi

# Generalized ESD Test for Outliers
# https://www.itl.nist.gov/div898/handbook/eda/section3/eda35h3.htm
def gesd():
    pass


def point_linear_interp(x, x_1, y_1, x_2, y_2):
    return (y_2 - y_1) / (x_2 - x_1) * (x - x_1) + y_1

"""
extrapolation methods:
    saturation
"""
def series_linear_interp(orig_time, orig_signal, desired_time, exterp="saturation"):
    result = []
    for item in desired_time:
        if item < orig_time[0]:
            if exterp == "saturation":
                result.append(orig_signal[0])
            elif exterp == "linear":
                result.append(point_linear_interp(item, orig_time[0], orig_signal[0], orig_time[1], orig_signal[1]))
        if item > orig_time[-1]:
            if exterp == "saturation":
                result.append(orig_signal[-1])
            elif exterp == "linear":
                result.append(point_linear_interp(item, orig_time[-2], orig_signal[-2], orig_time[-1], orig_signal[-1]))
        for i in range(len(orig_time) - 1):
            if orig_time[i] <= item and item <= orig_time[i + 1]:
                result.append(point_linear_interp(item, orig_time[i], orig_signal[i], orig_time[i + 1], orig_signal[i + 1]))
                break
    return result

def fill_null(time, signal):
    pass


from mathbox.calculus.differential import diff

def local_minmax(signal, closed=False):
    threshold = (max(signal) - min(signal)) * 0.001 # avoid floating point errors
    max_pos = []
    min_pos = []
    diff_x = diff(signal)
    gradient = [x if abs(x) > threshold else 0.0 for x in diff_x]
    
    if closed == True:
        if gradient[0] > 0:
            min_pos.append(0)
        elif gradient[0] < 0:
            max_pos.append(0)
        else:
            for i in range(1, len(gradient)):
                if gradient[i] < 0:
                    max_pos.extend(range(0, i + 1))
                    break
                elif gradient[i] > 0:
                    min_pos.extend(range(0, i + 1))
                    break
                else:
                    continue

    for i in range(len(gradient) - 1):
        if gradient[i] > 0 and gradient[i + 1] < 0:
            max_pos.append(i + 1)
            continue
        if gradient[i] < 0 and gradient[i + 1] > 0:
            min_pos.append(i + 1)
            continue
        # > < or < > continue
        if gradient[i] == 0:
            continue
        if gradient[i] > 0 and gradient[i + 1] == 0:
            for j in range(i + 1, len(gradient)):
                if gradient[j] < 0:
                    max_pos.extend(range(i + 1, j + 1))
                    break
                if gradient[j] == 0:
                    continue
                if gradient[j] > 0:
                    break
        if gradient[i] < 0 and gradient[i + 1] == 0:
            for j in range(i + 1, len(gradient)):
                if gradient[j] > 0:
                    min_pos.extend(range(i + 1, j + 1))
                    break
                if gradient[j] == 0:
                    continue
                if gradient[j] < 0:
                    break

    if closed == True:
        if gradient[-1] < 0:
            min_pos.append(len(gradient))
        elif gradient[-1] > 0:
            max_pos.append(len(gradient))
        else:
            for i in range(len(gradient) - 1, 0, -1):
                if gradient[i] > 0:
                    max_pos.extend(range(len(gradient), i, -1))
                    break
                elif gradient[i] < 0:
                    min_pos.extend(range(len(gradient), i, -1))
                    break
                else:
                    continue
    
    max_points = [(i,x) for i,x in enumerate(signal) if i in max_pos]
    min_points = [(i,x) for i,x in enumerate(signal) if i in min_pos]
    return max_points, min_points

def argmax(alist):
    # np.argmax output the 1st, this output the last
    return max((x,i) for i,x in enumerate(alist))[1]


from setuptools import find_packages, setup

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='mathbox',
    python_requires='>=3.8.0',
    version='0.0.8',
    description="A math toolbox.",
    author='freedeaths',
    author_email='register917@gmail.com',
    url='https://github.com/freedeaths/mathbox-py.git',
    #package_dir={'': 'mathbox'},
    packages=find_packages(exclude=["tests"]),
    #packages=['mathbox', 'mathbox/*'],
    install_requires=[],
    include_package_data=True,
    license="MIT",
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['math', 'data-processing'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
    ],
)


from mathbox.statistics.estimator import median
from mathbox.app.signal.transform import dftfreq, rdft, irdft
from mathbox.optimizer.utils import local_minmax

def moving_average(signal, window=3) :
    sum = 0
    result = [0 for _ in signal]
    for i in range(window):
        sum += signal[i]
        result[i] = sum / (i + 1)
    for i in range(window, len(signal)):
        sum += signal[i] - signal[i - window]
        result[i] = sum / window
    return result

def moving_median(signal, window=3):
    result = []
    for i in range(len(signal)):
        if i < window:
            result.append(median(signal[:i + 1]))
        else:
            result.append(median(signal[i - window + 1:i + 1]))
    return result

# fourier lowpass filter
# inputs:
#   timeseries: 1D timeseries signal
#   dt: sample period
#   n: number of the maximum amplitude
#   f_min: minimum frequency
# outputs:
#   freq: the frequencies whose amplitude is maximum
#   f_timeseries: filtered timeseries
def f_lowpass_filter(timeseries, dt, n=10, f_min=0.2):
    sample_num = len(timeseries)
    f = dftfreq(sample_num, dt)
    f_half = f[:((sample_num - 1) // 2) + 1]
    f_sig = rdft(timeseries)
    if f_sig[-1].imag < 1e-10:
        fix_flag = False
    else:
        fix_flag = True

    f_sig_amp = [abs(x) for x in f_sig]

    offset = f_sig_amp[0] / (sample_num)

    amp_max_pos = [i[0] for i in local_minmax(f_sig_amp)[0]]
    amp_max = [e for i, e in enumerate(f_sig_amp) if i in amp_max_pos]
    amp_desc_pos = [i[0] for i in sorted(enumerate(amp_max), key=lambda x:x[1], reverse=True)]
    
    if len(amp_desc_pos) > n:
        amp_desc_pos = amp_desc_pos[:n]
    amp_max_n_pos = [amp_max_pos[i] for i in amp_desc_pos]

    freq = [f_half[x] for x in amp_max_n_pos]
    for i in range(len(f_sig)):
        if i not in amp_max_n_pos:
            f_sig[i] = 0

    for i in range(len(f_half)):
        if f_half[i] < f_min:
            f_sig[i] = 0
    
    f_timeseries = irdft(f_sig, fix_flag)
    f_timeseries = [x + offset for x in f_timeseries]
    return (freq, f_timeseries)